searchNodes=[{"type":"module","title":"Cassandrax","doc":"Cassandrax is a Cassandra ORM built on top of Xandra and Ecto. Xandra provides the driver for communication with the Database, and Ecto provides the data mapping functionality as well as changesets to control data mutations.Cassandrax is heavily inspired by Triton and Ecto projects, if you have used any of those projects, you'll get up to speed in no time to use Cassandrax.Cassandrax is split into 3 components following the design choices made on Ecto:Cassandrax.Keyspace - keyspaces are the touchpoints with your Cassandra. Keyspaces provide the API for inserting, updating and deleting data from your tables, as well as querying the data already stored. Keyspaces need a cluster or a connection and a name.Cassandrax.Schema - schemas are used to map data fetched from Cassandra into an Elixir struct. They use Ecto.Schema under the hood and should require a @primary_key to be defined before the table definition.Cassandrax.Query - following Ecto design, they're written in Elixir syntax and are used to retrieve data from a Cassandra table. Queries are composable, even though they should remain as straight forward as possible. Unlike with relational databases, where you first model the data and their relationships and later you define how you should query that data, on Cassandra query design decisions are made when modeling the tables. For more information, please refer to Cassandra DocsNext we'll provide an overview of these components and how you'll use them to insert/change/fetch from/to Cassandra. Please check their corresponding module documentation for more in-depth description of the features available and options.KeyspacesCassandrax.Keyspace is a wrapper around the keyspace. Each keyspace contains one or multiple tables and belongs to a cluster. A keyspace can be defined like so:defmodule SomeKeyspace do use Cassandrax.Keyspace, cluster: SomeCluster, name: &quot;some_keyspace&quot; endAnd the configuration for SomeCluster must be in your application environment, usually defined in your config/config.exs:config :cassandrax, clusters: [SomeCluster] config :cassandrax, SomeCluster, protocol_version: :v4, nodes: [&quot;127.0.0.1:9042&quot;] username: &quot;cassandra&quot;, password: &quot;cassandra&quot;, # cassandrax accepts all options you&#39;d use in xandraCassandrax automatically picks up these configs to start the pool of connections for each cluster. Keep in mind that all keyspaces that belong to the same cluster will share the same pool of connections. If you need your keyspace to have its own connection pool, please refer to the Cassandrax.Keyspace specific documentation.SchemasSchemas are used for table definition. Here's an example:defmodule UserByEmail do use Cassandrax.Schema # needs to be defined *before* defining the schema @primary_key [:email] # table name is users_by_email. Notice we don&#39;t need to set the keyspace here table &quot;users_by_email&quot; do field :email, :string field :id, :integer field :username, :string end endCassandrax uses Ecto.Schema to define a struct with the schema fields:iex&gt; user_by_email = %UserByEmail{email: &quot;user@example.com&quot;} iex&gt; user_by_email.email &quot;user@example.com&quot;Just like with Ecto, this schema allows us to interact with keyspaces, like so:iex&gt; user_by_email = %UserByEmail{email: &quot;user@example.com&quot;, id: 123_456, username: &quot;user&quot;} iex&gt; SomeKeyspace.insert!(user_by_email) %UserByEmail{...}Unlike relational databases which come with the autoincrement ID as default primary key, Cassandra requires you to define your own primary key. Therefore calling Keyspace.insert/2 always returns the struct itself with updated metadata, but no changed fields.Also, bear in mind that Cassandra doesn't provice consistency guarantees the same way relational databases do, so the returned values of, for instance, deleting a record that doesn't exist anymore is the same as deleting an existing one:iex&gt; user_by_email = %UserByEmail{email: &quot;user@example.com&quot;, id: 123_456, username: &quot;user&quot;} # Store the result in a variable iex&gt; result = SomeKeyspace.insert!(user_by_email) %UserByEmail{...} # Now delete the recently inserted record... iex&gt; SomeKeyspace.delete!(result) %UserByEmail{...} # And if you try to delete a record that doesn&#39;t exist, no error is returned iex&gt; SomeKeyspace.delete!(result) %UserByEmail{...}QueriesCassandrax provides you with a DSL so you can write queries in Elixir, lowering the chances of writing invalid CQL statements. In some occasions, Cassandrax.Query will validate your query at compile time and fail as soon as possible if your query is invalid:import Ecto.Query query = UserByEmail |&gt; where(:email == &quot;user@example.com&quot;) # Returns a List of %UserByEmail{} structs matching the query result = SomeKeyspace.all(query)Specific examples and detailed documentation for all available keywords are available in Cassandrax.Query module docs, but the supported keywords are::allow_filtering:distinct:group_by:limit:order_by:per_partition_limit:select:whereCassandrax.Keyspace provides the same API as Ecto: You have Keyspace.all/1 which returns all records matching a query, Keyspace.one/1 which returns a single entry or raises and Keyspace.get/2 which fetches an entry by its primary key.","ref":"Cassandrax.html"},{"type":"function","title":"Cassandrax.cql/4","doc":"","ref":"Cassandrax.html#cql/4"},{"type":"function","title":"Cassandrax.ensure_cluster_config!/2","doc":"","ref":"Cassandrax.html#ensure_cluster_config!/2"},{"type":"function","title":"Cassandrax.start/2","doc":"Callback implementation for Application.start/2.","ref":"Cassandrax.html#start/2"},{"type":"function","title":"Cassandrax.start_link/1","doc":"","ref":"Cassandrax.html#start_link/1"},{"type":"function","title":"Cassandrax.stream_cql/4","doc":"","ref":"Cassandrax.html#stream_cql/4"},{"type":"module","title":"Cassandrax.Connection","doc":"","ref":"Cassandrax.Connection.html"},{"type":"function","title":"Cassandrax.Connection.all/2","doc":"","ref":"Cassandrax.Connection.html#all/2"},{"type":"function","title":"Cassandrax.Connection.child_spec/1","doc":"","ref":"Cassandrax.Connection.html#child_spec/1"},{"type":"function","title":"Cassandrax.Connection.delete/3","doc":"","ref":"Cassandrax.Connection.html#delete/3"},{"type":"function","title":"Cassandrax.Connection.execute/2","doc":"","ref":"Cassandrax.Connection.html#execute/2"},{"type":"function","title":"Cassandrax.Connection.execute/4","doc":"","ref":"Cassandrax.Connection.html#execute/4"},{"type":"function","title":"Cassandrax.Connection.insert/3","doc":"","ref":"Cassandrax.Connection.html#insert/3"},{"type":"function","title":"Cassandrax.Connection.prepare/2","doc":"","ref":"Cassandrax.Connection.html#prepare/2"},{"type":"function","title":"Cassandrax.Connection.stream_pages/4","doc":"","ref":"Cassandrax.Connection.html#stream_pages/4"},{"type":"function","title":"Cassandrax.Connection.update/4","doc":"","ref":"Cassandrax.Connection.html#update/4"},{"type":"behaviour","title":"Cassandrax.Keyspace","doc":"Defines a Keyspace.A keyspace acts as a repository, wrapping an underlying keyspace in CassandraDB.Setuptest_conn_attrs = [ nodes: [&quot;127.0.0.1:9043&quot;], username: &quot;cassandra&quot;, password: &quot;cassandra&quot; ]child = Cassandrax.Supervisor.child_spec(Cassandrax.MyConn, test_conn_attrs) Cassandrax.start_link([child])Defining a new keyspace module.defmodule MyKeyspace do use Cassandrax.Keyspace, cluster: Cassandrax.MyConn, name: &quot;my_keyspace&quot; endCreating a keyspace.statement = &quot;&quot;&quot; CREATE KEYSPACE IF NOT EXISTS my_keyspace WITH REPLICATION = {&#39;class&#39;: &#39;SimpleStrategy&#39;, &#39;replication_factor&#39;: 1} &quot;&quot;&quot; Cassandrax.cql(Cassandrax.MyConn, statement)Creating a table in the Keyspace.statement = [ &quot;CREATE TABLE IF NOT EXISTS &quot;, &quot;my_keyspace.user(&quot;, &quot;id integer, &quot;, &quot;user_name text, &quot;, &quot;svalue set&lt;text&gt;, &quot;, &quot;PRIMARY KEY (id))&quot; ] {:ok, _result} = Cassandrax.cql(Cassandrax.MyConn, statement)","ref":"Cassandrax.Keyspace.html"},{"type":"callback","title":"Cassandrax.Keyspace.__conn__/0","doc":"Accesses the cluster that was setup in the runtime configuration.ExampleCassandrax.MyConn = MyKeyspace.__conn__()","ref":"Cassandrax.Keyspace.html#c:__conn__/0"},{"type":"callback","title":"Cassandrax.Keyspace.__default_options__/1","doc":"Accesses the consistency level that manages availability versus data accuracy. Consistency level is configured for per individual read or write operation. Pass :read or :write to access the consistency level (eg. [consistency: :one]).Example[consistency: :one] = MyKeyspace.__default_options__(:read)","ref":"Cassandrax.Keyspace.html#c:__default_options__/1"},{"type":"callback","title":"Cassandrax.Keyspace.__keyspace__/0","doc":"Accesses the name of the Keyspace.Example&quot;my_keyspace&quot; = MyKeyspace.__keyspace__()","ref":"Cassandrax.Keyspace.html#c:__keyspace__/0"},{"type":"callback","title":"Cassandrax.Keyspace.all/2","doc":"Fetches all entries from the data store that matches the given query.May raise Xandra.Error if query validation fails.Examplequery = where(User, id: 1) MyKeyspace.all(query)","ref":"Cassandrax.Keyspace.html#c:all/2"},{"type":"callback","title":"Cassandrax.Keyspace.batch/2","doc":"Runs batch queries.Can be used to group and execute queries as Cassandra BATCH query.Options:logged is the default behavior in Cassandrax. Logged batch acts like a lightweight transaction around a batch operation. It enforces atomicity, and fails the batch if any of the queries fail. Cassandra doesn't enforce any other transactional properties at batch level.:unlogged consider it when there are multiple inserts and updates for the same partition key. Unlogged batching will give a warning if too many operations or too many partitions are involved.Read the CassandraDB documents for more information logged and unlogged batch operations.Exampleuser = MyKeyspace.get(User, id: 1) changeset = Ecto.Changeset.change(user, user_name: &quot;trent&quot;) MyKeyspace.batch(fn batch -&gt; batch |&gt; MyKeyspace.batch_insert(%User{id: 3, user_name: &quot;eve&quot;}) |&gt; MyKeyspace.batch_insert(%User{id: 4, user_name: &quot;mallory&quot;}) |&gt; MyKeyspace.batch_update(changeset) |&gt; MyKeyspace.batch_delete(user) end)","ref":"Cassandrax.Keyspace.html#c:batch/2"},{"type":"callback","title":"Cassandrax.Keyspace.batch_delete/2","doc":"Adds a DELETE query to the given batch.","ref":"Cassandrax.Keyspace.html#c:batch_delete/2"},{"type":"callback","title":"Cassandrax.Keyspace.batch_insert/2","doc":"Adds an INSERT query to the given batch.","ref":"Cassandrax.Keyspace.html#c:batch_insert/2"},{"type":"callback","title":"Cassandrax.Keyspace.batch_update/2","doc":"Adds an UPDATE query to the given batch.","ref":"Cassandrax.Keyspace.html#c:batch_update/2"},{"type":"callback","title":"Cassandrax.Keyspace.cql/3","doc":"Runs plain CQL Statements.Returns {:ok, map} if the CQL is successfully run or {:error, message} if there was a validation or a known constraint error.Examplestatement = &quot;&quot;&quot; SELECT * my_keyspace.user &quot;&quot;&quot; Cassandrax.cql(MyConn, statement)","ref":"Cassandrax.Keyspace.html#c:cql/3"},{"type":"callback","title":"Cassandrax.Keyspace.delete/2","doc":"Deletes a struct using its primary key.If the struct has no primary key, Xandra.Error will be raised.If the struct has been removed from db prior to call, it will still return {:ok, Cassandrax.Schema.t()}It returns {:ok, struct} if the struct has been successfully deleted or {:error, message} if there was a validation or a known constraint error.ExampleMyKeyspace.delete(%User(id: 1, user_name: &quot;bob&quot;))","ref":"Cassandrax.Keyspace.html#c:delete/2"},{"type":"callback","title":"Cassandrax.Keyspace.delete!/2","doc":"Same as delete/2 but returns the struct or raises if the changeset is invalid.","ref":"Cassandrax.Keyspace.html#c:delete!/2"},{"type":"callback","title":"Cassandrax.Keyspace.get/3","doc":"ExampleMyKeyspace.get(User, 2)","ref":"Cassandrax.Keyspace.html#c:get/3"},{"type":"callback","title":"Cassandrax.Keyspace.init/2","doc":"A callback executed when the keyspace starts or when configuration is read.The first argument is the context the callback is being invoked. If it is called because the Keyspace supervisor is starting, it will be :supervisor. It will be :runtime if it is called for reading configuration without actually starting a process.The second argument is the keyspace configuration as stored in the application environment. It must return {:ok, keyword} with the updated list of configuration or :ignore (only in the :supervisor case).","ref":"Cassandrax.Keyspace.html#c:init/2"},{"type":"callback","title":"Cassandrax.Keyspace.insert/2","doc":"Inserts a struct defined in Cassandrax.Schema or a changeset.If a struct is given, the struct is converted into a changeset with all non-nil fields.Example{:ok, user} = MyKeyspace.insert(%User{id: 1, user_name: &quot;bob&quot;})","ref":"Cassandrax.Keyspace.html#c:insert/2"},{"type":"callback","title":"Cassandrax.Keyspace.insert!/2","doc":"Same as insert/2 but returns the struct or raises if the changeset is invalid.","ref":"Cassandrax.Keyspace.html#c:insert!/2"},{"type":"callback","title":"Cassandrax.Keyspace.one/2","doc":"Fetches a single record from the query.Returns nil if no records were found. May raise Cassandrax.MultipleResultsError, if query returns more than one entry.Examplequery = where(User, id: 1) MyKeyspace.one(query)","ref":"Cassandrax.Keyspace.html#c:one/2"},{"type":"callback","title":"Cassandrax.Keyspace.stream/2","doc":"Streams all entries from the data store that matches the given query.May raise Xandra.Error if query validation fails.Examplequery = where(User, id: 1) MyKeyspace.stream(query)","ref":"Cassandrax.Keyspace.html#c:stream/2"},{"type":"callback","title":"Cassandrax.Keyspace.update/2","doc":"Updates a changeset using its primary key.Requires a changeset as it is the only way to track changes.If the struct has no primary key, Xandra.Error will be raised. In CassandraDB, UPDATE is also an upsert. If the struct cannot be found, a new entry will be created.It returns {:ok, struct} if the struct has been successfully updated or {:error, message} if there was a validation or a known constraint error.Exampleuser = MyKeyspace.get(User, 1) changeset = Ecto.Changeset.change(user, user_name: &quot;tom&quot;) MyKeyspace.update(changeset)","ref":"Cassandrax.Keyspace.html#c:update/2"},{"type":"callback","title":"Cassandrax.Keyspace.update!/2","doc":"Same as update/2 but returns the struct or raises if the changeset is invalid.","ref":"Cassandrax.Keyspace.html#c:update!/2"},{"type":"module","title":"Cassandrax.Keyspace.Schema","doc":"This module is invoked by User defined keyspaces for schema related functionality","ref":"Cassandrax.Keyspace.Schema.html"},{"type":"function","title":"Cassandrax.Keyspace.Schema.batch_delete/3","doc":"Implementation for Cassandrax.Keyspace.batch_delete/2","ref":"Cassandrax.Keyspace.Schema.html#batch_delete/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.batch_insert/3","doc":"Implementation for Cassandrax.Keyspace.batch_insert/2","ref":"Cassandrax.Keyspace.Schema.html#batch_insert/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.batch_update/3","doc":"Implementation for Cassandrax.Keyspace.batch_update/2","ref":"Cassandrax.Keyspace.Schema.html#batch_update/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.delete/3","doc":"Implementation for Cassandrax.Keyspace.delete/2.","ref":"Cassandrax.Keyspace.Schema.html#delete/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.delete!/3","doc":"Implementation for Cassandrax.Keyspace.delete!/2.","ref":"Cassandrax.Keyspace.Schema.html#delete!/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.insert/3","doc":"Implementation for Cassandrax.Keyspace.insert/2.","ref":"Cassandrax.Keyspace.Schema.html#insert/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.insert!/3","doc":"Implementation for Cassandrax.Keyspace.insert!/2.","ref":"Cassandrax.Keyspace.Schema.html#insert!/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.update/3","doc":"Implementation for Cassandrax.Keyspace.update/2.","ref":"Cassandrax.Keyspace.Schema.html#update/3"},{"type":"function","title":"Cassandrax.Keyspace.Schema.update!/3","doc":"Implementation for Cassandrax.Keyspace.update!/2.","ref":"Cassandrax.Keyspace.Schema.html#update!/3"},{"type":"behaviour","title":"Cassandrax.Query","doc":"Provides the query macros. Queries are used to retrieve or manipulate data from a repository (see Cassandrax.Keyspace).","ref":"Cassandrax.Query.html"},{"type":"macro","title":"Cassandrax.Query.allow_filtering/1","doc":"","ref":"Cassandrax.Query.html#allow_filtering/1"},{"type":"callback","title":"Cassandrax.Query.allow_filtering/2","doc":"A query expression that enables filtering in certain Cassandra queries.CassandraDB doesn't allow certain queries to be executed for performance reasons, such as where. You need to set ALLOW FILTERING to bypass this block. More details in CassandraDB docs.Examplequery = User |&gt; allow_filtering() |&gt; where(:id &gt; 1) |&gt; where(:user_name != &quot;alice&quot;)","ref":"Cassandrax.Query.html#c:allow_filtering/2"},{"type":"macro","title":"Cassandrax.Query.distinct/2","doc":"","ref":"Cassandrax.Query.html#distinct/2"},{"type":"callback","title":"Cassandrax.Query.distinct/2","doc":"A distinct query expression.Only returns the distinct records from the result. Only works with a list of partition_key(s).##Examplequery = distinct(TestSchema, [:id])","ref":"Cassandrax.Query.html#c:distinct/2"},{"type":"macro","title":"Cassandrax.Query.group_by/2","doc":"","ref":"Cassandrax.Query.html#group_by/2"},{"type":"callback","title":"Cassandrax.Query.group_by/2","doc":"A group by query expression.Allows to condense into a single row all selected rows that share the same values for a set of columns. Only available for partition key level or at a clustering column level.Examplequery = User |&gt; allow_filtering() |&gt; group_by([:id])","ref":"Cassandrax.Query.html#c:group_by/2"},{"type":"macro","title":"Cassandrax.Query.limit/2","doc":"","ref":"Cassandrax.Query.html#limit/2"},{"type":"callback","title":"Cassandrax.Query.limit/2","doc":"A limit query expression.Limits the number of rows to be returned from the result. Requires an integer, fields cannot be included. Default limit is 100.Limit expressions are chainable, however, the last limit expression will take precedence.Examplequery = limit(User, 200)","ref":"Cassandrax.Query.html#c:limit/2"},{"type":"macro","title":"Cassandrax.Query.order_by/2","doc":"","ref":"Cassandrax.Query.html#order_by/2"},{"type":"callback","title":"Cassandrax.Query.order_by/2","doc":"An order by query expression.Orders the fields based on a given key or list of keys. Order by needs to be paired with a where clause, specifically with where clauses that have equality or in. You also need to setup the table correctly to be able to perform order by queries.Example Table Setupstatement = [ &quot;CREATE TABLE IF NOT EXISTS &quot;, &quot;MyKeyspace.ordered_(&quot;, &quot;id int, &quot;, &quot;device_id int, &quot;, &quot;value text, &quot;, &quot;PRIMARY KEY (id, device_id))&quot;, &quot;WITH CLUSTERING ORDER BY (device_id DESC)&quot; ] Cassandrax.cql(MyConn, statement)Examplequery = User |&gt; allow_filtering() |&gt; where(:id == 1) |&gt; order_by([:device_id])You can set an explicit order.query = User |&gt; allow_filtering() |&gt; where(:id == 1) |&gt; order_by([asc: :device_id])","ref":"Cassandrax.Query.html#c:order_by/2"},{"type":"macro","title":"Cassandrax.Query.per_partition_limit/2","doc":"","ref":"Cassandrax.Query.html#per_partition_limit/2"},{"type":"callback","title":"Cassandrax.Query.per_partition_limit/2","doc":"A per partition limit expression controls the number of results return from each partition.Cassandra will then return only the first number of rows given in the per_partition_limit (clustered by the partition key) from that partition, regardless of how many ocurences of when may be present. More details in CassandraDB docs.ExampleDefault per_partition_limit is 100.query = per_partition_limit(User)Or you can set a custom per_partition_limitquery = per_partition_limit(User, 10)","ref":"Cassandrax.Query.html#c:per_partition_limit/2"},{"type":"macro","title":"Cassandrax.Query.select/2","doc":"","ref":"Cassandrax.Query.html#select/2"},{"type":"callback","title":"Cassandrax.Query.select/2","doc":"A select query expression.Selects the fields from the schema and any transformations that should be performed on the fields. Any expression that is accepted in a query can be a select field.Allows a list, tuple or a map. A full schema can also be selected. Only one select expression allowed in a query. If there is no select expression, the full schema will be selected by default.Accepts a list of atoms where atoms refer to fields.Examplequery = select(User, [:id]) %Cassandrax.Query{from: &quot;users&quot;, schema: Cassandrax.User, select: [:id]} = query","ref":"Cassandrax.Query.html#c:select/2"},{"type":"macro","title":"Cassandrax.Query.where/2","doc":"","ref":"Cassandrax.Query.html#where/2"},{"type":"callback","title":"Cassandrax.Query.where/2","doc":<<65, 32, 119, 104, 101, 114, 101, 32, 113, 117, 101, 114, 121, 32, 101, 120, 112, 114, 101, 115, 115, 105, 111, 110, 32, 116, 104, 97, 116, 32, 119, 111, 114, 107, 115, 32, 108, 105, 107, 101, 32, 97, 110, 32, 65, 78, 68, 32, 111, 112, ...>>,"ref":"Cassandrax.Query.html#c:where/2"},{"type":"type","title":"Cassandrax.Query.t/0","doc":"","ref":"Cassandrax.Query.html#t:t/0"},{"type":"module","title":"Cassandrax.Query.Builder","doc":"Builds query clauses and adds them to a Cassandrax.Query","ref":"Cassandrax.Query.Builder.html"},{"type":"function","title":"Cassandrax.Query.Builder.add_fragment/3","doc":"","ref":"Cassandrax.Query.Builder.html#add_fragment/3"},{"type":"function","title":"Cassandrax.Query.Builder.build/3","doc":"Converts the given data into a query clause and adds it to the given Cassandrax.Query.","ref":"Cassandrax.Query.Builder.html#build/3"},{"type":"function","title":"Cassandrax.Query.Builder.build_fragment/2","doc":"","ref":"Cassandrax.Query.Builder.html#build_fragment/2"},{"type":"protocol","title":"Cassandrax.Queryable","doc":"Converts a data structure into an Cassandrax.Query.Example defmodule User do use Cassandrax.Schema @primary_key [:id] table &quot;users&quot; do field :id, :integer end end %Cassandrax.Query{from: &quot;users&quot;, schema: User} = Cassandrax.Queryable.to_query(User)","ref":"Cassandrax.Queryable.html"},{"type":"function","title":"Cassandrax.Queryable.to_query/1","doc":"Converts the given data into an Cassandrax.Query.May raise a CompileError if the data is invalid.","ref":"Cassandrax.Queryable.html#to_query/1"},{"type":"type","title":"Cassandrax.Queryable.t/0","doc":"All the types that implement this protocol.","ref":"Cassandrax.Queryable.html#t:t/0"},{"type":"behaviour","title":"Cassandrax.Schema","doc":"Defines a schema.This schema is used to map data fetched from a CassandraDB node into an Elixir struct.Cassandrax.Schema mixin uses Ecto.Schema mixin.","ref":"Cassandrax.Schema.html"},{"type":"callback","title":"Cassandrax.Schema.convert/1","doc":"Converts a map of data into a struct for this module.","ref":"Cassandrax.Schema.html#c:convert/1"},{"type":"macro","title":"Cassandrax.Schema.table/2","doc":"Defines an embedded schema for the Cassandra table with the given field definitions.In order to create a schema, you must define a @primary_key before the schema definition. Unlike Ecto.Schema, Cassandrax.Schema won't automatically generate a primary key which is named id. @primary_key configures the schema primary key and it expects a list of key(s). You can set a single primary key which is the partition key in Cassandra or a list of keys where the first key is the partition key and the rest are the clustering keys which are responsible for sorting data within the partition.You can use Ecto's schema to leverage field definitions and metadata.Example defmodule User do use Cassandrax.Schema @primary_key [:id] table &quot;users&quot; do field :id, :integer field :user_name, :string end end","ref":"Cassandrax.Schema.html#table/2"},{"type":"type","title":"Cassandrax.Schema.t/0","doc":"","ref":"Cassandrax.Schema.html#t:t/0"},{"type":"module","title":"Cassandrax.Schema.MapSetType","doc":"Defines the SET collection type used by CassandraDB","ref":"Cassandrax.Schema.MapSetType.html"},{"type":"function","title":"Cassandrax.Schema.MapSetType.cast/1","doc":"Callback implementation for Ecto.Type.cast/1.","ref":"Cassandrax.Schema.MapSetType.html#cast/1"},{"type":"function","title":"Cassandrax.Schema.MapSetType.dump/1","doc":"Callback implementation for Ecto.Type.dump/1.","ref":"Cassandrax.Schema.MapSetType.html#dump/1"},{"type":"function","title":"Cassandrax.Schema.MapSetType.embed_as/1","doc":"Callback implementation for Ecto.Type.embed_as/1.","ref":"Cassandrax.Schema.MapSetType.html#embed_as/1"},{"type":"function","title":"Cassandrax.Schema.MapSetType.equal?/2","doc":"Callback implementation for Ecto.Type.equal?/2.","ref":"Cassandrax.Schema.MapSetType.html#equal?/2"},{"type":"function","title":"Cassandrax.Schema.MapSetType.load/1","doc":"Callback implementation for Ecto.Type.load/1.","ref":"Cassandrax.Schema.MapSetType.html#load/1"},{"type":"function","title":"Cassandrax.Schema.MapSetType.type/0","doc":"Callback implementation for Ecto.Type.type/0.","ref":"Cassandrax.Schema.MapSetType.html#type/0"},{"type":"exception","title":"Cassandrax.ClusterConfigError","doc":"Raised at runtime when a cluster is listed in the env, but no config for it is found","ref":"Cassandrax.ClusterConfigError.html"},{"type":"exception","title":"Cassandrax.ConnectionError","doc":"Raised at startup when a cluster cannot connect, to prevent application initialization when cassandrax isn't available","ref":"Cassandrax.ConnectionError.html"},{"type":"exception","title":"Cassandrax.MultipleResultsError","doc":"Raised at runtime when a query expects one result, but many are returned","ref":"Cassandrax.MultipleResultsError.html"},{"type":"exception","title":"Cassandrax.QueryError","doc":"Raised at runtime when a query is invalid","ref":"Cassandrax.QueryError.html"},{"type":"exception","title":"Cassandrax.SchemaError","doc":"Raised at compile time when a schema is invalid","ref":"Cassandrax.SchemaError.html"},{"type":"extras","title":"Cassandrax","doc":"Cassandrax Cassandrax is a Cassandra data mapping toolkit built on top of Ecto and query builder and runner on top of Xandra.Cassandrax is heavily inspired by the Triton and Ecto projects. It allows you to build and run CQL statements as well as map results to Elixir structs.The docs can be found at https://hexdocs.pm/cassandrax.","ref":"readme.html"},{"type":"extras","title":"Cassandrax - Installation","doc":"def deps do [ {:cassandrax, &quot;~&gt; 0.1.0&quot;} ] end","ref":"readme.html#installation"},{"type":"extras","title":"Cassandrax - Setup","doc":"test_conn_attrs = [ nodes: [&quot;127.0.0.1:9043&quot;], username: &quot;cassandra&quot;, password: &quot;cassandra&quot; ] # MyApp.MyCluster is just an atom child = Cassandrax.Supervisor.child_spec(MyApp.MyCluster, test_conn_attrs) Cassandrax.start_link([child])Alternatively, if you're using CassandraDB on a Phoenix app, you can edit your config/config.exs file to add Cassandrax to your supervision tree:# In your config/config.exs, you can add as many clusters as you like config :cassandrax, clusters: [MyApp.MyCluster] config :cassandrax, MyApp.MyCluster, protocol_version: :v4, nodes: [&quot;127.0.0.1:9042&quot;], pool_size: System.get_env(&quot;CASSANDRADB_POOL_SIZE&quot;) || 10, username: System.get_env(&quot;CASSANDRADB_USER&quot;) || &quot;cassandra&quot;, password: System.get_env(&quot;CASSANDRADB_PASSWORD&quot;) || &quot;cassandra&quot;, # Default write/read options write_options: [consistency: :local_quorum], read_options: [consistency: :one]","ref":"readme.html#setup"},{"type":"extras","title":"Cassandrax - Usage","doc":"You can easily define a Keyspace module that will act as a wrapper for read/write operations:defmodule MyKeyspace do use Cassandrax.Keyspace, cluster: MyApp.MyCluster, name: &quot;my_keyspace&quot; endTo define your schema, use the Cassandrax.Schema module, which provides the table macro:defmodule UserById do use Cassandrax.Schema # Defines :id as partition key and :age as clustering key @primary_key [:id, :age] table &quot;user_by_id&quot; do field :id, :integer field :age, :integer field :user_name, :string field :nicknames, MapSetType end endWhile we work to support an actual migration DSL, you can run plain CQL statements to migrate the database schema, like so:iex(1)&gt; statement = &quot;&quot;&quot; CREATE KEYSPACE IF NOT EXISTS my_keyspace WITH REPLICATION = {&#39;class&#39;: &#39;SimpleStrategy&#39;, &#39;replication_factor&#39;: 1} &quot;&quot;&quot; # Creating the Keyspace iex(2)&gt; Cassandrax.cql(MyApp.MyCluster, statement) {:ok, %Xandra.SchemaChange{ effect: &quot;CREATED&quot;, options: %{keyspace: &quot;my_keyspace&quot;}, target: &quot;KEYSPACE&quot;, tracing_id: nil }} iex(3)&gt; statement = &quot;&quot;&quot; CREATE TABLE IF NOT EXISTS my_keyspace.user_by_id( id int, age int, user_name varchar, nicknames set&lt;varchar&gt;, PRIMARY KEY (id, age)) &quot;&quot;&quot; # Creating the Table iex(4)&gt; Cassandrax.cql(MyApp.MyCluster, statement) {:ok, %Xandra.SchemaChange{ effect: &quot;CREATED&quot;, options: %{keyspace: &quot;my_keyspace&quot;, subject: &quot;user_by_id&quot;}, target: &quot;TABLE&quot;, tracing_id: nil }}MigrationsIn future, we plan to support pure cassandrax migrations, but so far we still depend on Ecto to keep track of migrations. Below we present a strategy to keep cassandrax migrations separated from your main database migrations.Let's configure a new Ecto.Repo to put migrations on priv/cassandrax_repo/migrations:# Configure an additional Ecto.Repo config :my_app, MyApp.CassandraxRepo, database: &quot;same as your main database&quot;, hostname: &quot;localhost&quot;, username: &quot;username&quot;, password: &quot;password&quot; config :my_app, MyApp.CassandraxRepo, # ensure cassandrax connection is ready before the migration runs start_apps_before_migration: [:cassandrax],Then create the additional Ecto.Repo pointing to a different table than schema_migrations, to not conflict with your main database migrations.defmodule MyApp.CassandraxRepo do @moduledoc &quot;&quot;&quot; Keep track of versions for Cassandra migrations. &quot;&quot;&quot; use Ecto.Repo, otp_app: :repo, adapter: Ecto.Adapters.Postgres, migration_source: &quot;cassandra_migrations&quot; endNow you can simply create a new migration withmix ecto.gen.migration -r MyApp.CassandraxRepo create_first_table`And edit the filedefmodule Repo.Migrations.CreateFirstTable do use Ecto.Migration alias MyApp.MyCluster def up do statement = &quot;&quot;&quot; CREATE TABLE IF NOT EXISTS my_keyspace.user_by_id( id int, age int, user_name varchar, nicknames set&lt;varchar&gt;, PRIMARY KEY (id, age)) &quot;&quot;&quot; {:ok, _result} = Cassandrax.cql(Cluster, statement) end def down do statement = &quot;DROP TABLE IF EXISTS my_keyspace.user_by_id&quot; {:ok, _result} = Cassandrax.cql(Cluster, statement) end endAlso, remember to include MyApp.CassandraxRepo migrations on your deploy scripts!CRUDMutating data is as easy as it is with a regular Ecto schema. You can work straight with structs, or with changesets:Insertiex(5)&gt; user = %UserById{id: 1, user_name: &quot;alice&quot;} %UserById{id: 1, user_name: &quot;alice&quot;} iex(6)&gt; MyKeyspace.insert(user) {:ok, %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;alice&quot;}} iex(7)&gt; MyKeyspace.insert!(user) %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;alice&quot;}Updateiex(8)&gt; changeset = Changeset.change(user, user_name: &quot;bob&quot;) #Ecto.Changeset&lt;changes: %{user_name: &quot;bob&quot;}, ...&gt; iex(9)&gt; MyKeyspace.update(changeset) {:ok, %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;bob&quot;}} iex(10)&gt; MyKeyspace.update!(changeset) %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;bob&quot;}Deleteiex(11)&gt; MyKeyspace.delete(user) {:ok, %UserById{__meta__: %Ecto.Schema.Metadata{:deleted, &quot;user_by_id&quot;}, id: 1, user_name: &quot;bob&quot;}} iex(12)&gt; MyKeyspace.delete!(user) %UserById{__meta__: %Ecto.Schema.Metadata{:deleted, &quot;user_by_id&quot;}, id: 1, user_name: &quot;bob&quot;}Batch operationsYou can issue many operation at once with a BATCH operation. For more information on how Batches work in Cassandra DB, please refer to CassandraDB Batches.iex(13)&gt; user = %UserById{id: 1, user_name: &quot;alice&quot;} %UserById{id: 1, user_name: &quot;alice&quot;} iex(14)&gt; changeset = MyKeyspace.get(UserById, id: 2) |&gt; Changeset.change(user_name: &quot;eve&quot;) #Ecto.Changeset&lt;changes: %{user_name: &quot;eve&quot;, ...}&gt; iex(15)&gt; MyKeyspace.batch(fn batch -&gt; batch |&gt; MyKeyspace.batch_insert(user) |&gt; MyKeyspace.batch_update(changeset) end) :okQueryingDisclaimerOne thing to keep in mind when it comes to querying is the API is still under development and, therefore, can still change in version prior to 0.1.0.If you use it in production, be cautious when updating cassandrax, and make sure all your queries work correctly after installing the new version.Cassandrax queries are very similar to Ecto's, you can use the all/2, get/2 and one/2 functions directly from your Keyspace module.iex(16)&gt; MyKeyspace.get(UserById, [id: 1, age: 20]) %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, age: 20, user_name: &quot;alice&quot;} iex(17)&gt; MyKeyspace.all(UserById) [ %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;alice&quot;}, %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 2, user_name: &quot;eve&quot;}, ... ]Also, you can use Cassandrax.Query macros to build your own queries.iex(18)&gt; import Cassandrax.Query true iex(19)&gt; UserById |&gt; where(id: 1, age: 20) |&gt; MyKeyspace.one() %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, age: 20, user_name: &quot;alice&quot;} # Remember when filtering data by non-primary key fields, you should use ALLOW FILTERING: iex(20)&gt; UserById |&gt; where(id: 3) |&gt; where(:user_name == &quot;adam&quot;) |&gt; where(:age &gt;= 30) |&gt; allow_filtering() |&gt; MyKeyspace.all() [%UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 3, user_name: &quot;adam&quot;, age: 31}}]Streaming data is supported.iex(21)&gt; users = MyKeyspace.stream(UserById, page_size: 20) #Function&lt;59.58486609/2 in Stream.transform/3&gt; iex(22) Emum.to_list(users) [ %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 1, user_name: &quot;alice&quot;}, %UserById{__meta__: %Ecto.Schema.Metadata{:loaded, &quot;user_by_id&quot;}, id: 2, user_name: &quot;eve&quot;}, ... ]","ref":"readme.html#usage"}]